%{ // -*- C++ -*-

// yacc/bison grammar for Esterel

// Stephen Edwards

#include <suifkernel/module.h>
#include <basicnodes/basic.h>
#include <basicnodes/basic_factory.h>
#include <esuifnodes/esuif.h>
#include <esuifnodes/esuif_factory.h>
#include <common/lstring.h>

// The bison-generated parser uses malloc() and free(), defined in stdlib.h
#include <stdlib.h>

// The bison-generated parse uses strcat()
#include <string.h>

#include <stdio.h>

// Make the error messages more interesting.  YYDEBUG must be non-zero for
// YYERROR_VERBOSE to work.  Debugging is not fully enabled unless yydebug = 1

#define YYDEBUG 1
#define YYERROR_VERBOSE

extern int yydebug;
extern int yylineno;

extern FILE *yyin;

// FileBlock currently being parsed
static FileBlock *file_block;

// SuifEnv currently being used
static SuifEnv *suif_env;

void yyerror(const char *s)
{
  fprintf(stderr, "%s\n", s);
}

extern int yylex();

%}

%{ // Token definitions
%}

%token ABORT AND AWAIT CALL CASE COMBINE CONSTANT COPYMODULE DO EACH ELSE
%token ELSIF EMIT END EVERY EXIT FALSE FUNCTION HALT HANDLE IF IMMEDIATE
%token IN INPUT INPUTOUTPUT LOOP MOD MODULE NOT NOTHING OR OUTPUT PAUSE
%token POSITIVE PRE PRESENT PROCEDURE RELATION REPEAT RETURN RUN SENSOR SIGNAL
%token SUSPEND SUSTAIN THEN TIMEOUT TIMES TRAP TRUE TYPE UPTO VAR WATCHING
%token WEAK WHEN WITH

%token INTEGER BOOLEAN STRINGCONST

%token COLEQ IMPLIES TWOQUEST PARALLEL

%token <i> INTLIT
%token <s> ID

%token '+' '-' '*' '/' '<' '>' GE LE '.' '=' NE

%{ // Precedence relations
%}

%left OR
%left AND
%left NOT
%left '<' '>' LE GE '=' NE
%left '+' '-'
%left '*' '/' MOD
%right UnaryNegation

%left ';'


%union{
  const char *s;
  int i;
}

%%

// Modules

File :
    ModuleList { @1.first_line; }
  ;

// This right-recursive definition avoids a shift/reduce conflict.

ModuleList :
    Module
  | Module MODULE
  | Module ModuleList
  | Module MODULE ModuleList 
  ;

Module :
    MODULE ID ':' InterfaceDeclListOpt Statement END
      {
	ModuleSymbol *ms = create_module_symbol(suif_env);
        ms->set_name($2);
        file_block->get_symbol_table()->append_symbol_table_object(ms);
        ModuleDefinition *md = create_module_definition(suif_env);
	md->set_procedure_symbol(ms);
	file_block->get_definition_block()->append_procedure_definition(md);
      }
  ;

// Interface Declaration

InterfaceDeclListOpt :
    { /* Empty */ }
  | InterfaceDeclList
  ;

InterfaceDeclList :
    InterfaceDecls
  | InterfaceDeclList InterfaceDecls
  ;

InterfaceDecls :
    TypeDecls
  | ConstantDecls
  | FunctionDecls
  | ProcedureDecls
/*  | TaskDecls */
  | InterfaceSignalDecls
  | SensorDecls
  | RelationDecls
/*  | OutputDecls */ /* Missing from the primer grammar. */
  ;

// Type Declarations

TypeDecls :
  TYPE TypeDeclList ';'

TypeDeclList :
    TypeDecl
  | TypeDeclList ',' TypeDecl

TypeDecl :
    ID {}
  ;

// Constant Declarations

ConstantDecls :
    CONSTANT ConstantDeclList ';'
  ;

ConstantDeclList :
    ConstantDecl
  | ConstantDeclList ',' ConstantDecl
  ;

ConstantDecl :
    ID ':' TypeIdentifier {}
  | ID '=' ConstantAtom ':' TypeIdentifier {}
  ;

// Function Declarations

FunctionDecls :
    FUNCTION FunctionDeclList ';'

FunctionDeclList :
    FunctionDecl
  | FunctionDeclList ',' FunctionDecl
  ;

FunctionDecl :
    ID '(' IdentifierListOpt ')' ':' TypeIdentifier {}
  ;

IdentifierListOpt :
    { /* Empty */ }
  | IdentifierList
  ;

IdentifierList :
    ID {}
  | IdentifierList ',' ID {}
  ;

// Procedure Declarations

ProcedureDecls :
    PROCEDURE ProcedureDeclList ';'
  ;

ProcedureDeclList :
    ProcedureDecl
  | ProcedureDeclList ',' ProcedureDecl

ProcedureDecl :
    ID '(' IdentifierListOpt ')' '(' IdentifierListOpt ')' {}
  ;

// Task Declarations

/* Not supported */

// Signal Declarations

InterfaceSignalDecls :
    INPUT SignalDeclList ';'
  | OUTPUT SignalDeclList ';'
  | INPUTOUTPUT SignalDeclList ';'
  | RETURN SignalDeclList ';'
  ;

SignalDeclList :
    SignalDecl
  | SignalDeclList ',' SignalDecl
  ;

SignalDecl :
    ID {}
  | ID ':' ChannelType {}
  | ID COLEQ Expression ':' ChannelType {}
  ;

ChannelTypeOpt :
    { /* Empty */ }
  | ChannelType
  ;

ChannelType :
    TypeIdentifier
  | COMBINE TypeIdentifier WITH FunctionIdentifier
  | COMBINE TypeIdentifier WITH PredefinedCombineFunction
  ;

PredefinedCombineFunction :
    '+'
  | '*'
  | OR
  | AND
  ;

// Sensor Declarations

SensorDecls :
    SENSOR SensorDeclList ';'
  ;

SensorDeclList :
    SensorDecl
  | SensorDeclList ',' SensorDecl
  ;

SensorDecl : 
    ID ':' TypeIdentifier {}
  ;

// Input Relation Declarations

RelationDecls :
    RELATION RelationDeclList ';'
  ;

RelationDeclList :
    RelationDecl
  | RelationDeclList ',' RelationDecl
  ;

RelationDecl :
    ImplicationDecl
  | ExclusionDecl
  ;

ImplicationDecl :
    SignalIdentifier IMPLIES SignalIdentifier
  ;

ExclusionDecl :
    SignalIdentifier '#' SignalIdentifier
  | ExclusionDecl '#' SignalIdentifier
  ;

// Expressions

// Data Expressions

Expression :
    Constant
  | '(' Expression ')'
  | '?' SignalIdentifier
  | PRE '(' '?' SignalIdentifier ')'
  | TWOQUEST ExceptionIdentifier
  | UnaryNegation '-' Expression
  | Expression '*' Expression
  | Expression '/' Expression
  | Expression '+' Expression
  | Expression '-' Expression
  | Expression MOD Expression
  | Expression '=' Expression
  | Expression NE Expression
  | Expression '<' Expression
  | Expression LE Expression
  | Expression '>' Expression
  | Expression GE Expression
  | NOT Expression
  | Expression AND Expression
  | Expression OR Expression
  | FunctionCall
  ;

Constant :
    ConstantLiteral
  | UnsignedNumber
  ;

ConstantLiteral :
    ConstantIdentifier
  | TRUE
  | FALSE
  | StringConstant
  ;

ConstantAtom :
    ConstantLiteral
  | SignedNumber
  ;

SignedNumber :
    UnsignedNumber
  | '-' UnsignedNumber
  ;

UnsignedNumber : 
    UnsignedIntegerConstant
/*  | UnsignedFloatConstant */ /* FIXME */
/*  | UnsignedDoubleConstant */ /* FIXME */
  ;

FunctionCall :
    FunctionIdentifier '(' ExpressionListOpt ')'
  ;

ExpressionListOpt :
    { /* Empty */ }
  | ExpressionList
  ;

ExpressionList :
    Expression
  | ExpressionList ',' Expression
  ;

// Signal Expressions

SignalExpression :
    SignalIdentifier
  | PRE '(' SignalIdentifier ')'
  | NOT SignalExpression
  | SignalExpression AND SignalExpression
  | SignalExpression OR SignalExpression
  | '(' SignalExpression ')'
  ;

// Delay Expressions

DelayExpression :
    BracketedSignalExpression
  | IMMEDIATE BracketedSignalExpression
  | Expression BracketedSignalExpression
  ;
  
BracketedSignalExpression :
    SignalIdentifier
  | '[' SignalExpression ']'
  ;

// Statements

Statement :
    Parallel
  | NonParallel
  ;

Parallel :
    NonParallel PARALLEL NonParallel
  | Parallel PARALLEL NonParallel
  ;

NonParallel :
    AtomicStatement
  | Sequence
  ;

Sequence :
    SequenceWithoutTerminator semiOpt

semiOpt :
    { /* Empty */ }
  | ';'
  ;

SequenceWithoutTerminator :
    AtomicStatement ';' AtomicStatement
  | SequenceWithoutTerminator ';' AtomicStatement
  ;

AtomicStatement :
    NOTHING
  | PAUSE
  | HALT
  | Emit
  | Sustain
  | Assignment
  | ProcedureCall
  | Present
  | If
  | Loop
  | Repeat
  | Abort
  | Await
  | LoopEach
  | Every
  | Suspend
  | Trap
  | Exit
/* | Exec */
  | LocalVariableDecl
  | LocalSignalDecl
  | RunModule
  | '[' Statement ']'
  | DoWatching
  | DoUpto
  ;

// Signal Emission

Emit :
    EMIT SignalIdentifier
  | EMIT SignalIdentifier '(' Expression ')'
  ;

Sustain :
    SUSTAIN SignalIdentifier
  | SUSTAIN SignalIdentifier '(' Expression ')'
  ;

// Assignment and Procedure Call

Assignment :
    VariableIdentifier COLEQ Expression
  ;

ProcedureCall :
    CALL ProcedureIdentifier '(' IdentifierListOpt ')'
                             '(' ExpressionListOpt ')'
  ;

// The present Signal Test

Present :
    PresentThenElse
  | PresentCaseElse
  ;

PresentThenElse :
    /* PRESENT PresentEvent ThenPartOpt ElsePartOpt END PRESENTopt */
    PRESENT PresentEvent ThenPart ElsePartOpt END PRESENTopt
  | PRESENT PresentEvent ElsePartOpt END PRESENTopt
  ;

PRESENTopt :
    { /* Empty */ }
  | PRESENT
  ;

PresentEvent :
    SignalExpression
  | '[' SignalExpression ']'
  ;

ThenPartOpt :
    { /* Empty */ }
  | ThenPart
  ;

ThenPart :
    THEN Statement
  ;

ElsePartOpt :
    { /* Empty */ }  
  |  ELSE Statement
  ;

PresentCaseElse :
    PRESENT PresentCaseList ElsePartOpt END PRESENTopt
  ;

PresentCaseList :
    PresentCase
  | PresentCaseList PresentCase
  ;

PresentCase :
    CASE PresentEvent
  | CASE PresentEvent DO Statement
  ;

// The if Data Test

If :
    IF Expression ThenPartOpt ElsifPartListOpt ElsePartOpt END IFopt
  ;

IFopt :
    { /* Empty */ }
  | IF
  ;

ElsifPartListOpt :
    { /* Empty */ }
  | ElsifPartList
  ;

ElsifPartList :
    Elsif
  | ElsifPartList Elsif
  ;

Elsif :
    ELSIF Expression ThenPartOpt
  ;

// Looping

Loop :
    LOOP Statement END LOOPopt
  ;

LOOPopt :
    { /* Empty */ }
  | LOOP
  ;

// Repeat Loops

Repeat :
    POSITIVEopt REPEAT Expression TIMES Statement END REPEATopt
  ;

POSITIVEopt :
    { /* Empty */ }
  | POSITIVE
  ;

REPEATopt :    
    { /* Empty */ }
  | REPEAT
  ;

// The abort Statements

Abort :
    WEAKopt ABORT Statement WHEN DelayExpression
  | WEAKopt ABORT Statement WHEN DelayExpression
    DO Statement END WEAKopt ABORTopt
  | WEAKopt ABORT Statement WHEN AbortCaseList END WEAKopt ABORTopt
  ;

WEAKopt :
    { /* Empty */ }
  | WEAK
  ;

ABORTopt:
    { /* Empty */ }
  | ABORT
  ;

AbortCaseList :
    AbortCase
  | AbortCaseList AbortCase
  ;

AbortCase :
    CASE DelayExpression DO Statement
  | CASE DelayExpression
  ;

// The await Statement

Await :
    AWAIT DelayExpression
  | AWAIT DelayExpression DO Statement END AWAITopt
  | AWAIT AwaitCaseList END AWAITopt
  ;

AWAITopt :
    { /* Empty */ }
  | AWAIT
  ;

AwaitCaseList :
    AwaitCase
  | AwaitCaseList AwaitCase
  ;

AwaitCase :
    CASE DelayExpression DO Statement
  | CASE DelayExpression
  ;

// Temporal Loops

LoopEach :
    LOOP Statement EACH DelayExpression
  ;

Every :
    EVERY DelayExpression DO Statement END EVERYopt
  ;

EVERYopt :
    { /* Empty */ }
  | EVERY
  ;

// The suspend Statement

Suspend : 
    SUSPEND Statement WHEN DelayExpression
  ;

// Traps

Trap :
    TRAP ExceptionDeclList IN Statement ExceptionHandlerListOpt END TRAPopt
  ;

TRAPopt :
    { /* Empty */ }
  | TRAP
  ;

ExceptionDeclList :
    ExceptionDecl
  | ExceptionDeclList ',' ExceptionDecl
  ;

ExceptionDecl :
    ID ChannelTypeOpt {}
  | ID ':' ChannelType {}
  | ID COLEQ Expression ':' ChannelType {}
  ;

Exit :
    EXIT ID {}
  | EXIT ID '(' Expression ')' {}

ExceptionHandlerListOpt:
    { /* Empty */ }
  | ExceptionHandlerList
  ;

ExceptionHandlerList :
    ExceptionHandler
  | ExceptionHandlerList ExceptionHandler
  ;

ExceptionHandler :
    HANDLE ExceptionEvent DO Statement
  ;

ExceptionEvent :
    ID {}
  | '(' ExceptionEvent ')'
  | NOT ExceptionEvent
  | ExceptionEvent AND ExceptionEvent
  | ExceptionEvent OR ExceptionEvent
  ;

// The exec Task Execution Statement

/* Not Supported */

// Local Signal Declaration

LocalSignalDecl :
    SIGNAL SignalDeclList IN Statement END SIGNALopt
  ;

SIGNALopt :
    { /* Empty */ }
  | SIGNAL
  ;

// Local Variable Declaration

LocalVariableDecl :
    VAR VariableDeclList IN Statement END VARopt
  ;

VARopt :
    { /* Empty */ }
  | VAR
  ;

VariableDeclList :
    VariableDecl
  | VariableDeclList ',' VariableDecl
  ;

VariableDecl :
    ID ':' TypeIdentifier {}
  | ID COLEQ Expression ':' TypeIdentifier {}
  ;

// The run Module Instantiation Statement

RunModule :
    RUN RunModuleNames
  | RUN RunModuleNames '[' RenamingList ']'
  ;

RunModuleNames :
    ModuleIdentifier
  | ModuleIdentifier '/' ModuleIdentifier
  ;

RenamingList :
    Renaming
  | RenamingList ';' Renaming
  ;

Renaming :
    TYPE TypeRenamingList
  | CONSTANT ConstantRenamingList
  | FUNCTION FunctionRenamingList
  | PROCEDURE ProcedureRenamingList
/*  | TASK TaskRenamingList */
  | SIGNAL SignalRenamingList
  ;

TypeRenamingList :
    TypeRenaming
  | TypeRenamingList ',' TypeRenaming
  ;

TypeRenaming :
    TypeIdentifier '/' TypeIdentifier
  ;

ConstantRenamingList :
    ConstantRenaming
  | ConstantRenamingList ',' ConstantRenaming
  ;

ConstantRenaming :
    ConstantAtom '/' ConstantIdentifier
  ;

FunctionRenamingList :
    FunctionRenaming
  | FunctionRenamingList ',' FunctionRenaming
  ;

FunctionRenaming :
    FunctionIdentifier '/' FunctionIdentifier
  | PredefinedFunction '/' FunctionIdentifier
  ;

PredefinedFunction :
    AND
  | OR
  | NOT
  | '+'
  | '-'
  | '*'
  | '/'
  | MOD
  | '<'
  | '>'
  | LE
  | GE
  | NE
  | '='
  ;

ProcedureRenamingList :
    ProcedureRenaming
  | ProcedureRenamingList ',' ProcedureRenaming
  ;

ProcedureRenaming :
    ProcedureIdentifier '/' ProcedureIdentifier
  ;

SignalRenamingList :
    SignalRenaming
  | SignalRenamingList ',' SignalRenaming
  ;

SignalRenaming :
    SignalIdentifier '/' SignalIdentifier
  ;

// Deprecated Syntax

DoWatching :
    DO Statement WATCHING DelayExpression
  | DO Statement WATCHING DelayExpression TIMEOUT Statement END TIMEOUTopt
  ;

TIMEOUTopt :
    { /* Empty */ }
  | TIMEOUT
  ;

DoUpto :
    DO Statement UPTO DelayExpression
  ;



TypeIdentifier : ID {} ;
SignalIdentifier : ID {} ;
ExceptionIdentifier : ID {} ;
ConstantIdentifier : ID {} ;
FunctionIdentifier : ID {} ;
VariableIdentifier : ID {} ;
ProcedureIdentifier : ID {} ;
ModuleIdentifier : ID {} ;

StringConstant : STRINGCONST {} ;
UnsignedIntegerConstant : INTLIT {} ;
UnsignedFloatConstant : INTLIT {} ;
UnsignedDoubleConstant : INTLIT {} ;


%%

void est2s_parse_file(const char *filename, FileSetBlock *fsb, SuifEnv *env)
{
  FILE *file;
  file = fopen(filename, "r");
  if (!file) {
    fprintf(stderr, "Could not open %s\n", filename);
    return;
  }
  yyin = file;
  yydebug = 1;

  printf("Parsing %s\n", filename);
  suif_env = env;
  file_block = create_file_block(suif_env, filename);
  yyparse();
  fsb->append_file_block(file_block);
  printf("Finished parsing %s\n", filename);
  fclose(file);
}
